import numpy as np
import torch
import torch.nn as nn
import os

import anyact as aa

import matplotlib.pyplot as plt
plt.ion()
plt.ioff()

device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

lossfunc2 = aa.MSECosLoss()

def trainer(mult,batch_size=2**14,epochs=2**12,lri=1e-7,lr0=-1,lr1=-2, training_σ=1, lossfunc=nn.MSELoss()):
	mult = mult.to(device)
	lrs = 10**(torch.linspace(lr0,lr1,epochs))
	# start slow to avoid overshooting on first step to crazy values... seems to not be working
	# intro_epochs = int(epochs*0.1)
	# lrs = 10**(torch.cat([torch.linspace(lri,lr0,intro_epochs),torch.linspace(lr0,lr1,epochs-intro_epochs)]))
	optim = torch.optim.Adam(mult.parameters(),lr=1e-2,eps=1e-12)
	hist = []
	for epoch in range(epochs):
		mult.train()
		lr = (lrs[epoch]).item()
		for g in optim.param_groups: g['lr'] = lr
		### many matrices
		A = torch.randn(batch_size,*(mult.dim_A),device=device)*training_σ # 10
		B = torch.randn(batch_size,*(mult.dim_B),device=device)*training_σ # 10
		### matrix multiplication
		y = A@B # ground truth
		Y = mult(A,B)
		### training
		loss = lossfunc(y,Y)
		# loss = nn.MSELoss()(y.repeat(batch_size,1,1,1),Y) # this will trade the warning for CUDA OUT OF MEMORY
		# loss = nn.MSELoss()(y.expand(batch_size, *y.shape),Y) # same problem :(
		optim.zero_grad()
		loss.backward()
		optim.step()
		print(f"{epoch:5d} | loss={loss.item():5.8f}  log(loss)={np.log(loss.item()):5.8f}")
		hist.append(loss.item())
	# if plot: plt.scatter(range(len(hist)),np.log(hist),s=.05,label=name)
	mult.hist = hist
	return mult

# def viz_error(mult,batch_size=2**14):
# 		A = torch.randn(batch_size,*(mult.dim_A),device=device)
# 		B = torch.randn(batch_size,*(mult.dim_B),device=device)
# 		y = A@B # = torch.einsum('nij,njk->nik',A,B) # ground truth
# 		Y = mult(A,B)
# 		plt.hist(torch.flatten(torch.log((y-Y)**2)).detach().cpu(),128,density=True)
# 		# plt.hist(torch.flatten(((y-Y)**2)).detach().cpu(),128,density=True)

def viz_error(name,mult,batch_size=2**16):
	os.makedirs(f"out/extra/",exist_ok=True)
	A = torch.randn(batch_size,*(mult.dim_A),device=device)
	B = torch.randn(batch_size,*(mult.dim_B),device=device)
	truth = A@B
	found = mult(A,B)
	error = ((truth-found)**2).sum(dim=2).sum(dim=1).sqrt()
	error = error.log().clamp(-10,10)
	error = error.cpu().detach()
	plt.close("all");plt.figure(figsize=[3.2,2.4])
	plt.hist(error,512,density=True)
	# plt.xscale("log")
	plt.xlabel("log(error)")
	plt.ylabel("distribution")
	plt.tight_layout()
	plt.savefig(f"out/extra/{name}_error.png")
	# plt.show()
	###
	###
	###
	# nn.MSELoss()(mult2d(A,B),A@B).item()
	# dot = (truth*truth).sum(2).sum(1) / ((truth**2).sum(2).sum(1)) # debug : must be 1
	dot = (truth*found).sum(2).sum(1) / ((truth**2).sum(2).sum(1))
	dot = dot.cpu().detach()
	dot=dot.clamp(-10,10)
	plt.close("all");plt.figure(figsize=[3.2,2.4])
	plt.hist(dot,512,density=True)
	plt.plot([1,1],[0,1])
	# plt.plot([0,0],[0,1])
	plt.yscale("log")
	plt.grid()
	# plt.show()
	plt.tight_layout()
	plt.savefig(f"out/extra/{name}_dot.png")
	###
	# plt.xlim([-10,10])
	# plt.tight_layout()
	# plt.savefig(f"out/extra/{name}_dot_zoom.png")
	print(name, sum(dot<0)/len(dot))
	###
	###
	###
	B1 = torch.randn(batch_size,*(mult.dim_B),device=device)
	B2 = torch.randn(batch_size,*(mult.dim_B),device=device)
	direct = mult(A,B1+B2)
	indirect = mult(A,B1)+mult(A,B2)
	truth = A@(B1+B2)
	dir_indir = torch.log(torch.sqrt(((direct-indirect)**2).sum(2).sum(1))).clamp(-10,10)
	truth_dir = torch.log(torch.sqrt(((truth-direct)**2).sum(2).sum(1))).clamp(-10,10)
	truth_indir = torch.log(torch.sqrt(((truth-indirect)**2).sum(2).sum(1))).clamp(-10,10)
	# plt.close("all");plt.figure(figsize=[3.2,2.4])
	plt.close("all");plt.figure()
	plt.hist(dir_indir  .cpu().detach(),256,density=True,label="log([model(A,B1+B2)] - [model(A,B1)+model(A,B2)])",alpha=0.5)
	plt.hist(truth_dir  .cpu().detach(),256,density=True,label="log([A⋅(B1+B2)] - [model(A,B1+B2)])",alpha=0.5)
	plt.hist(truth_indir.cpu().detach(),256,density=True,label="log([A⋅(B1+B2)] - [model(A,B1)+model(A,B2)])",alpha=0.5)
	plt.legend(loc="lower left")
	# plt.xlabel("log([model(A,B1+B2)] - [model(A,B1)+model(A,B2)])")
	# plt.plot([1,1],[0,1])
	# # plt.plot([0,0],[0,1])
	plt.yscale("log")
	# plt.grid()
	# plt.show()
	plt.tight_layout()
	plt.savefig(f"out/extra/{name}_propt.png")


###############################################################################
### MODELS ####################################################################
###############################################################################

class MatMul(nn.Module):
	def __init__(self,dim_A,dim_B,depth=5,width=1):
		super().__init__()
		if not dim_A[1]==dim_B[0]: print("ERROR INCOMPATIBILITY DETECTED")
		self.dim_A = dim_A
		self.dim_B = dim_B
		self.flatten = nn.Flatten()
		dim_in  = dim_A[0]*dim_A[1] + dim_B[0]*dim_B[1]
		dim_mid = width*dim_in
		dim_out = dim_A[0]*dim_B[1]
		# self.model = nn.Sequential(*([
		# 	# ] + (depth>0)*[ nn.Linear(dim_in ,dim_mid)  , aa.Stack([nn.Identity(), nn.Sequential(nn.GELU(),nn.BatchNorm1d(dim_mid))]), # nn.BatchNorm1d(2*dim_mid),
		# 	# ] + (depth-1)*[ nn.Linear(2*dim_mid,dim_mid), aa.Stack([nn.Identity(), nn.Sequential(nn.GELU(),nn.BatchNorm1d(dim_mid))]), # nn.BatchNorm1d(2*dim_mid),
		# 	] + (depth>0)*[ nn.Linear(dim_in ,dim_mid)  , aa.Stack([nn.Identity(), nn.GELU()]), # nn.BatchNorm1d(2*dim_mid),
		# 	] + (depth-1)*[ nn.Linear(2*dim_mid,dim_mid), aa.Stack([nn.Identity(), nn.GELU()]), # nn.BatchNorm1d(2*dim_mid),
		# 	] + (depth>0)*[ nn.Linear(2*dim_mid,dim_out),
		# 	] + (depth<1)*[ nn.Linear(dim_in ,dim_out),
		# 	]
		# ))
		seq=[]
			# ] + (depth>0)*[ nn.Linear(dim_in ,dim_mid)  , aa.Stack([nn.Identity(), nn.Sequential(nn.GELU(),nn.BatchNorm1d(dim_mid))]), # nn.BatchNorm1d(2*dim_mid),
			# ] + (depth-1)*[ nn.Linear(2*dim_mid,dim_mid), aa.Stack([nn.Identity(), nn.Sequential(nn.GELU(),nn.BatchNorm1d(dim_mid))]), # nn.BatchNorm1d(2*dim_mid),
		if depth>0:
			seq.append( nn.Linear(dim_in ,dim_mid) )
			seq.append( aa.Stack([nn.Identity(), nn.GELU()]) )
			# seq.append( nn.BatchNorm1d(2*dim_mid) )
		for d in range(depth-1):
			seq.append( nn.Linear(2*dim_mid,dim_mid) )
			seq.append( aa.Stack([nn.Identity(), nn.GELU()]) )
			# seq.append( nn.BatchNorm1d(2*dim_mid) )
		if depth>0:
			seq.append( nn.Linear(2*dim_mid,dim_out) )
		if depth<1:
			seq.append( nn.Linear(dim_in ,dim_out) )
		self.model = nn.Sequential(*seq)
	def forward(self,A,B):
		AB = torch.hstack((self.flatten(A),self.flatten(B)))
		return self.model(AB).view(-1,self.dim_A[0],self.dim_B[1])

####################################
####################################
####################################
lri=1e-4
lr0=-1
lr1=-2
####################################
####################################
####################################
# training_σ=3
training_σ=5
####################################
####################################
####################################
dim=1
dim_A=[dim,dim]
dim_B=[dim,dim]
batch_size=2**15
epochs = 2**12
lr0=-2
lr1=-4
os.makedirs(f"out/",exist_ok=True)

def helper1a(widths,depths):
	mults=[]
	for depth in depths:
		for width in widths:
			name = f"depth={depth}, width={width}"
			mult = MatMul(dim_A,dim_B,depth=depth,width=width)
			print(name)
			print(f"parameter count = {sum([np.round(np.exp(sum(np.log(p.shape)))) for p in mult.parameters()])}")
			mult = trainer(mult,batch_size=batch_size,epochs=epochs,lr0=lr0,lr1=lr1, training_σ=training_σ)
			name = f"deep_scalar_depth={depth}_width={width}"
			mults.append([name,mult])
			# plt.scatter(range(len(mult.hist)),np.log(mult.hist),s=.2,label=name)
	return mults

def helper1b(mults):
	plt.clf()
	for [name,mult] in mults:
		plt.scatter(range(len(mult.hist)),np.log(mult.hist),s=.1,label=name)
	plt.xscale('log')
	plt.ylabel("log(MSE)")
	plt.xlabel("epoch")
	legend = plt.legend()
	for h in legend.legendHandles: h._sizes[0] = 6

widths=[8]
depths=[1,2,4,8,16]
mults_depths = helper1a(widths,depths)
helper1b(mults_depths)
# plt.title("scalar multiplication by depth")
plt.tight_layout()
plt.savefig(f"out/deep_scalar_by_depth_σ={training_σ}.png")
for [name,mult] in mults_depths: viz_error(name,mult)

widths=[1,2,4,8,16]
depths=[5]
mults_widths = helper1a(widths,depths)
helper1b(mults_widths)
# plt.title("scalar multiplication by width")
plt.tight_layout()
plt.savefig(f"out/deep_scalar_by_width_σ={training_σ}.png")
for [name,mult] in mults_widths: viz_error(name,mult)

widths=[16]
depths=[1,2,4,8,16]
mults_depths2 = helper1a(widths,depths)
helper1b(mults_depths2)
# plt.title("scalar multiplication by depth")
plt.tight_layout()
plt.savefig(f"out/deep_scalar_by_depth_w={widths[0]}_σ={training_σ}.png")
for [name,mult] in mults_depths: viz_error(name,mult)

widths=[1,2,4,8,16]
depths=[16]
mults_widths = helper1a(widths,depths)
helper1b(mults_widths)
# plt.title("scalar multiplication by width")
plt.tight_layout()
plt.savefig(f"out/deep_scalar_by_width_d={depths[0]}_σ={training_σ}.png")
for [name,mult] in mults_widths: viz_error(name,mult)

####################################
####################################
####################################
# # with torch.autograd.profiler.profile() as prof:
# with profile(activities=[ProfilerActivity.CPU,ProfilerActivity.CUDA], record_shapes=True) as prof:
# 	trainer(mult,batch_size=batch_size,epochs=2**5,training_σ=10)

# print(prof.key_averages().table(sort_by="self_cpu_time_total"))
####################################
####################################
####################################
dim=1
dim_A=[dim,dim]
dim_B=[dim,dim]
batch_size=2**15
epochs = 2**12 # epochs = 2**15
lr0=-2
lr1=-4 # lr1=-6
depth=8
width=8
mult = MatMul(dim_A,dim_B,depth=depth,width=width)
mult = trainer(mult,batch_size=batch_size,epochs=epochs,lr0=lr0,lr1=lr1, training_σ=training_σ)
name = f"deep_scalar_depth={depth}_width={width}"
plt.clf()
plt.scatter(range(len(mult.hist)),np.log(mult.hist),s=.1,label=name)
plt.xscale('log')
plt.ylabel("log(MSE)")
plt.xlabel("epoch")
# plt.legend()
plt.tight_layout()
plt.savefig(f"out/{name}_loss.png")

from PIL import Image

def error_mult(name,mult, x_range=[-100,100],y_range=[-100,100]):
	# x_range=[-60,60]
	# y_range=[-60,60]
	resolution = 2**10
	x_span = torch.linspace(x_range[0],x_range[1],resolution,device=device)
	y_span = torch.linspace(y_range[0],y_range[1],resolution,device=device)
	#
	x,y = torch.meshgrid(x_span,y_span)
	x = x.reshape(-1,1,1) # .reshape(2**11,2**11)
	y = y.reshape(-1,1,1) # .reshape(2**11,2**11)
	sgn       = ((mult(x,y)*(x*y) > 0)*2 - 1  ).reshape(resolution,resolution).detach().cpu().numpy()
	err_add   = ((mult(x,y) - (x*y))          ).reshape(resolution,resolution).detach().cpu().numpy()
	err_ratio = ((mult(x,y) / (x*y))-1        ).reshape(resolution,resolution).detach().cpu().numpy()
	# xy = torch.log(mult(x,y)) - torch.log(x*y)
	# sgn       = sgn      .reshape(resolution,resolution).detach().cpu().numpy()
	# err_add   = err_add  .reshape(resolution,resolution).detach().cpu().numpy()
	# err_ratio = err_ratio.reshape(resolution,resolution).detach().cpu().numpy()
	# rgb = np.stack((err_ratio*(err_ratio>0),np.sin(3.115*3*err_ratio)>0.9,-err_ratio*(err_ratio<0)),-1)
	rgb = np.stack((err_ratio*(err_ratio>0),err_ratio*0,-err_ratio*(err_ratio<0)),-1)
	Image.fromarray((rgb.clip(0,1)*255).astype(np.uint8)).save(f"out/{name}_error_ratio.png")
	rgb = np.stack((err_add  *(err_add  >0),err_add  *0,-err_add  *(err_add  <0)),-1)
	Image.fromarray((rgb.clip(0,1)*255).astype(np.uint8)).save(f"out/{name}_error_addition.png")
	rgb = np.stack((sgn      *(sgn      >0),sgn      *0,-sgn      *(sgn      <0)),-1)
	Image.fromarray((rgb.clip(0,1)*255).astype(np.uint8)).save(f"out/{name}_error_sgn.png")
	# plt.imshow(rgb.transpose(1,0,2),extent=(x_range[0],x_range[1],y_range[0],y_range[1]),origin="lower")
	# return Image.fromarray((rgb.clip(0,1)*255).astype(np.uint8))
	# plt.contour(np.abs( np.exp(z*np.cos(θ)*np.cos(θ/2)) ), alpha=0.5)

x_range = 40 # x_range = 70
error_mult(name,mult, x_range=[-x_range,x_range],y_range=[-x_range,x_range])

x_range = 25 # x_range = 70

resolution = 2**10
for z in [0.0001,0.001,0.01,0.1,1,10,100]:
	x = torch.linspace(-x_range,x_range,resolution,device=device).reshape(-1,1,1)
	y = z*torch.ones(x.shape,device=device)
	# plt.clf()
	plt.close("all");plt.figure(figsize=[3.2,2.4])
	plt.plot(x.reshape(-1).cpu(),(x*y).reshape(-1).detach().cpu(),label=f"x⋅{z}")
	plt.plot(x.reshape(-1).cpu(),mult(x,y).reshape(-1).detach().cpu(),label=f"model(x,{z})")
	plt.plot(x.reshape(-1).cpu(),mult(y,x).reshape(-1).detach().cpu(),label=f"model({z},x)")
	plt.legend()
	plt.tight_layout()
	plt.savefig(f"out/{name}_perf_x_{z}.png")

for z in [0.0001,0.001,0.01,0.1,1,10,100]:
	x = torch.linspace(-x_range,x_range,resolution,device=device).reshape(-1,1,1)
	y = z*torch.ones(x.shape,device=device)
	# plt.clf()
	plt.close("all");plt.figure(figsize=[3.2,2.4])
	plt.plot(x.reshape(-1).cpu(),z*torch.ones(x.shape).reshape(-1).detach().cpu(),label=f"{z}")
	plt.plot(x.reshape(-1).cpu(),mult(x,z/x).reshape(-1).detach().cpu(),label=f"model(x,{z}/x)")
	plt.plot(x.reshape(-1).cpu(),mult(z/x,x).reshape(-1).detach().cpu(),label=f"model({z}/x,x)")
	plt.legend()
	plt.tight_layout()
	plt.ylim([z*1.5+5,z*0.5-5])
	plt.savefig(f"out/{name}_perf_x_x_{z}.png")

for z in [0.0001,0.001,0.01,0.1,1,10,100]:
	x = torch.linspace(-x_range,x_range,resolution,device=device).reshape(-1,1,1)
	# plt.clf()
	# plt.close("all");plt.figure(figsize=[3.2,2.4])
	# plt.plot(x.reshape(-1).cpu(),-(x*x).reshape(-1).detach().cpu(),label=f"-x⋅x")
	# plt.plot(x.reshape(-1).cpu(),+(x*x).reshape(-1).detach().cpu(),label=f"x⋅x")
	# plt.plot(x.reshape(-1).cpu(),mult(x,-x).reshape(-1).detach().cpu(),label=f"model(x,-x)")
	# plt.plot(x.reshape(-1).cpu(),mult(-x,x).reshape(-1).detach().cpu(),label=f"model(-x,x)")
	# plt.savefig(f"out/{name}_perf_x2.png")
	plt.close("all");plt.figure(figsize=[3.2,2.4])
	plt.plot(x.reshape(-1).cpu(),z*(x*x).reshape(-1).detach().cpu(),label=f"{z}⋅x⋅x")
	plt.plot(x.reshape(-1).cpu(),mult(x,z*x).reshape(-1).detach().cpu(),label=f"model(x,{z}⋅x)")
	plt.plot(x.reshape(-1).cpu(),mult(z*x,x).reshape(-1).detach().cpu(),label=f"model({z}⋅x,x)")
	plt.legend()
	plt.tight_layout()
	plt.savefig(f"out/{name}_perf_x2_{z}.png")

# plt.plot(x.reshape(-1).cpu(),(x*x).reshape(-1).detach().cpu())
# plt.plot(x.reshape(-1).cpu(),mult(x,x).reshape(-1).detach().cpu())

# plt.plot(x.reshape(-1).cpu(),(-x*x).reshape(-1).detach().cpu())
# plt.plot(x.reshape(-1).cpu(),mult(-x,x).reshape(-1).detach().cpu())
# plt.plot(x.reshape(-1).cpu(),mult(x,-x).reshape(-1).detach().cpu())

####################################
####################################
####################################
def helper2(dim,width=8,depth=8,lossfunc=nn.MSELoss(),name_extra=""):
	dim_A=[dim,dim]
	dim_B=[dim,dim]
	 # depth=4
	lr0=-2
	lr1=-5
	epochs=2**12
	batch_size=2**8
	# batch_size=2**14
	name = f"deep_mat{dim}mat{dim}_depth={depth}_width={width}{name_extra}"
	mult = MatMul(dim_A,dim_B,depth=depth,width=width)
	mult = trainer(mult,batch_size=batch_size,epochs=epochs,lr0=lr0,lr1=lr1,lossfunc=lossfunc)
	plt.close("all");plt.figure()
	plt.scatter(range(len(mult.hist)),np.log(mult.hist),s=.2,label=name)
	plt.xscale('log')
	plt.ylabel(f"log(MSE)")
	plt.xlabel(f"epoch")
	# plt.title(f"mat2 multiplication")
	plt.savefig(f"out/{name}_loss.png")
	viz_error(name,mult,batch_size=2**14)
	return name,mult

name2,mult2 = helper2(2)

# name3,mult3 = helper2(3)
name3,mult3 = helper2(3,depth=6)

name4,mult4 = helper2(4,depth=4)

name5,mult5 = helper2(5,depth=3,width=10)

# depth=4
# width=8
# batch_size=2**10
name6,mult6 = helper2(6,depth=3,width=8)

name7,mult7 = helper2(7,depth=2,width=8)

name8,mult8 = helper2(8,depth=2,width=8)

name9,mult9 = helper2(9,depth=1,width=16)

plt.close("all");plt.figure()
plt.scatter(range(len(mult2.hist)),np.log(mult2.hist),s=.2,label=name2)
plt.scatter(range(len(mult3.hist)),np.log(mult3.hist),s=.2,label=name3)
plt.scatter(range(len(mult4.hist)),np.log(mult4.hist),s=.2,label=name4)
plt.scatter(range(len(mult5.hist)),np.log(mult5.hist),s=.2,label=name5)
plt.scatter(range(len(mult6.hist)),np.log(mult6.hist),s=.2,label=name6)
plt.scatter(range(len(mult7.hist)),np.log(mult7.hist),s=.2,label=name7)
plt.scatter(range(len(mult8.hist)),np.log(mult8.hist),s=.2,label=name8)
plt.scatter(range(len(mult9.hist)),np.log(mult9.hist),s=.2,label=name9)
plt.xscale('log')
plt.ylabel(f"log(MSE)")
plt.xlabel(f"epoch")
# plt.title(f"mat2 multiplication")
plt.legend()
plt.tight_layout()
plt.savefig(f"out/deep_mat23456_loss.png")


# def helper_error(name,mult,batch_size=2**16):
# 	os.makedirs(f"out/extra/",exist_ok=True)
# 	A = torch.randn(batch_size,*(mult.dim_A),device=device)
# 	B = torch.randn(batch_size,*(mult.dim_B),device=device)
# 	truth = A@B
# 	found = mult(A,B)
# 	dot = (truth*found).sum(2).sum(1) / ((truth**2).sum(2).sum(1))
# 	dot = dot.cpu().detach()
# 	dot=dot.clamp(-10,10)
# 	plt.hist(dot,512,density=True,alpha=0.5,label=name)
# 	plt.yscale("log")

def helper_error(name,mult,batch_size=2**16):
	os.makedirs(f"out/extra/",exist_ok=True)
	A = torch.randn(batch_size,*(mult.dim_A),device=device)
	B = torch.randn(batch_size,*(mult.dim_B),device=device)
	truth = A@B
	found = mult(A,B)
	dot = (truth*found).sum(2).sum(1) / ((truth**2).sum(2).sum(1))
	dot = dot.cpu().detach()
	dot=dot.clamp(-10,10)
	# plt.hist(dot,512,density=True,alpha=0.5,label=name)
	hihi = np.histogram(dot,64,density=True)
	plt.plot([(a+b)/2 for a,b in zip(hihi[1][1:],hihi[1][:-1])], hihi[0], label=name)
	plt.yscale("log")

# plt.close("all");plt.figure(figsize=[3.2,2.4])
plt.close("all");plt.figure()
helper_error(name2,mult2)
helper_error(name3,mult3)
helper_error(name4,mult4)
helper_error(name5,mult5)
helper_error(name6,mult6)
helper_error(name7,mult7)
helper_error(name8,mult8)
helper_error(name9,mult9)
plt.plot([1,1],[0,1])
plt.grid()
plt.legend(loc="lower left")
plt.tight_layout()
plt.savefig(f"out/deep_mat23456_dot.png")

####################################
####################################
####################################

name16,mult16 = helper2(16,depth=1,width=8)
# name24,mult24 = helper2(24,depth=1,width=8)
# name32,mult32 = helper2(32,depth=1,width=8)

# from types import SimpleNamespace
# mult16b = SimpleNamespace()
# mult16b.hist=mult16.hist

plt.close("all");plt.figure()
plt.scatter(range(len(mult16.hist)),np.log(mult16.hist),s=.2,label=name16)
# plt.scatter(range(len(mult32.hist)),np.log(mult32.hist),s=.2,label=name32)
plt.xscale('log')
plt.ylabel(f"log(MSE)")
plt.xlabel(f"epoch")
plt.legend()
plt.tight_layout()
# plt.savefig(f"out/deep_mat9999_loss.png")
plt.savefig(f"out/{name16}_loss.png")

plt.close("all");plt.figure()
helper_error(name16,mult16)
# helper_error(name32,mult32)
plt.plot([1,1],[0,1])
plt.grid()
plt.legend(loc="lower left")
plt.tight_layout()
# plt.savefig(f"out/deep_mat9999_dot.png")
plt.savefig(f"out/{name16}_dot.png")
# viz_error(name16,mult16,batch_size=2**13)

####################################
####################################
####################################

name2,mult2 = helper2(2,lossfunc=lossfunc2,name_extra="_loss=msecosloss")

# name3,mult3 = helper2(3)
name3,mult3 = helper2(3,depth=6,lossfunc=lossfunc2,name_extra="_loss=msecosloss")

name4,mult4 = helper2(4,depth=4,lossfunc=lossfunc2,name_extra="_loss=msecosloss")

name5,mult5 = helper2(5,depth=3,width=10,lossfunc=lossfunc2,name_extra="_loss=msecosloss")

# depth=4
# width=8
# batch_size=2**10
name6,mult6 = helper2(6,depth=3,width=8,lossfunc=lossfunc2,name_extra="_loss=msecosloss")

name7,mult7 = helper2(7,depth=2,width=8,lossfunc=lossfunc2,name_extra="_loss=msecosloss")

name8,mult8 = helper2(8,depth=2,width=8,lossfunc=lossfunc2,name_extra="_loss=msecosloss")

name9,mult9 = helper2(9,depth=1,width=16,lossfunc=lossfunc2,name_extra="_loss=msecosloss")

plt.close("all");plt.figure()
plt.scatter(range(len(mult2.hist)),np.log(mult2.hist),s=.2,label=name2)
plt.scatter(range(len(mult3.hist)),np.log(mult3.hist),s=.2,label=name3)
plt.scatter(range(len(mult4.hist)),np.log(mult4.hist),s=.2,label=name4)
plt.scatter(range(len(mult5.hist)),np.log(mult5.hist),s=.2,label=name5)
plt.scatter(range(len(mult6.hist)),np.log(mult6.hist),s=.2,label=name6)
plt.scatter(range(len(mult7.hist)),np.log(mult7.hist),s=.2,label=name7)
plt.scatter(range(len(mult8.hist)),np.log(mult8.hist),s=.2,label=name8)
plt.scatter(range(len(mult9.hist)),np.log(mult9.hist),s=.2,label=name9)
plt.xscale('log')
plt.ylabel(f"log(MSE)")
plt.xlabel(f"epoch")
# plt.title(f"mat2 multiplication")
plt.legend()
plt.tight_layout()
plt.savefig(f"out/deep_mat23456_loss=msecossloss_loss.png")

# plt.close("all");plt.figure(figsize=[3.2,2.4])
plt.close("all");plt.figure()
helper_error(name2,mult2)
helper_error(name3,mult3)
helper_error(name4,mult4)
helper_error(name5,mult5)
helper_error(name6,mult6)
helper_error(name7,mult7)
helper_error(name8,mult8)
helper_error(name9,mult9)
plt.plot([1,1],[0,1])
plt.grid()
plt.legend(loc="lower left")
plt.tight_layout()
plt.savefig(f"out/deep_mat23456_loss=msecossloss_dot.png")


####################################
####################################
####################################

def helper3(dim_in,dim_out,depth=8,width=8,	lr0=-2,lr1=-5,batch_size=2**12):
	dim_A=[dim_in,dim_out]
	dim_B=[dim_out,1]
	# depth=5
	epochs=2**13
	name = f"deep_mat{dim_in}x{dim_out}_depth={depth}_width={width}"
	mult = MatMul(dim_A,dim_B,depth=depth,width=width)
	plt.close("all");plt.figure()
	mult = trainer(mult,batch_size=batch_size,epochs=epochs,lr0=lr0,lr1=lr1)
	plt.scatter(range(len(mult.hist)),np.log(mult.hist),s=.05,label=name)
	plt.xscale('log')
	plt.ylabel(f"log(MSE)")
	plt.xlabel(f"epoch")
	plt.title(f"mat2⋅vec2")
	plt.legend()
	plt.grid()
	plt.savefig(f"out/{name}_loss.png")
	viz_error(name,mult,batch_size=2**14)
	return name,mult

name2,mult2 = helper3(2,2)
name3,mult3 = helper3(3,3)
name4,mult4 = helper3(4,4,depth=6)

# depth=5
# lr0=-3
# name16,mult16 = helper3(16,16,depth=4,batch_size=2**10)
name16,mult16 = helper3(16,16,depth=4)


# lr0=-4
# lr1=-6
name32,mult32 = helper3(32,32,depth=2,width=2,batch_size=2**10)

# name32,mult32 = helper3(16,16,depth=4,batch_size=2**8,lr1=-3)







plt.close("all");plt.figure()
plt.scatter(range(len(mult2.hist)),np.log(mult2.hist),s=.2,label=name2)
plt.scatter(range(len(mult3.hist)),np.log(mult3.hist),s=.2,label=name3)
plt.scatter(range(len(mult4.hist)),np.log(mult4.hist),s=.2,label=name4)
plt.scatter(range(len(mult16.hist)),np.log(mult16.hist),s=.2,label=name16)
plt.scatter(range(len(mult32.hist)),np.log(mult32.hist),s=.2,label=name32)
plt.xscale('log')
plt.ylabel(f"log(MSE)")
plt.xlabel(f"epoch")
# plt.title(f"mat2 multiplication")
plt.legend()
plt.tight_layout()
plt.savefig(f"out/deep_matvec_loss.png")


# plt.close("all");plt.figure(figsize=[3.2,2.4])
plt.close("all");plt.figure()
helper_error(name2,mult2)
helper_error(name3,mult3)
helper_error(name4,mult4)
helper_error(name16,mult16)
helper_error(name32,mult32,batch_size=2**15)
plt.plot([1,1],[0,1])
plt.grid()
plt.legend(loc="lower left")
plt.tight_layout()
plt.savefig(f"out/deep_matvec_dot.png")



####################################
####################################
####################################

####################################
####################################
####################################









































































from complex import C_Log, Exp, ToReal, ToComplex
from linear import *

os.makedirs(f"out/",exist_ok=True)

# plt.clf()
# dim=1
# mult = MatMul([dim,dim],[dim,dim],depth=0,width=4);trainer_square(mult,dim,N=2**12,epochs=2**11)
# mult = MatMul([dim,dim],[dim,dim],depth=1,width=4);trainer_square(mult,dim,N=2**12,epochs=2**11)
# mult = MatMul([dim,dim],[dim,dim],depth=2,width=4);trainer_square(mult,dim,N=2**12,epochs=2**11)
# mult = MatMul([dim,dim],[dim,dim],depth=3,width=4);trainer_square(mult,dim,N=2**12,epochs=2**11)
# mult = MatMul([dim,dim],[dim,dim],depth=4,width=4);trainer_square(mult,dim,N=2**12,epochs=2**11)
# mult = MatMul([dim,dim],[dim,dim],depth=8,width=4);trainer_square(mult,dim,N=2**12,epochs=2**11)

class Mul_Log(nn.Module):
	# def __init__(self):
	# 	super().__init__()
	# 	self.model = nn.Sequential(
	# 		ToComplex(), C_Log(),
	# 		C_Linear_Binary_Sigmoid(2,1),
	# 		Exp(), ToReal(),
	# 	)
	def __init__(self,dim_A,dim_B,width=1,log_mixer_mode="binary",out_mixer_mode="2softmax"):
		super().__init__()
		if not dim_A[1]==dim_B[0]: print("ERROR INCOMPATIBILITY DETECTED")
		self.dim_A = dim_A
		self.dim_B = dim_B
		self.width=width
		self.flatten = nn.Flatten()
		dim_in  = dim_A[0]*dim_A[1] + dim_B[0]*dim_B[1]
		dim_mid = int(width*dim_in)
		dim_out = dim_A[0]*dim_B[1]
		if log_mixer_mode=="linear"    : log_mixer = C_Linear(dim_in,dim_mid,bias=False)
		if log_mixer_mode=="2softmax"  : log_mixer = C_Linear_2_Softmax(dim_in,dim_mid)
		if log_mixer_mode=="3softmax"  : log_mixer = C_Linear_3_Softmax(dim_in,dim_mid)
		if log_mixer_mode=="binary_sig": log_mixer = C_Linear_Binary_Sigmoid(dim_in,dim_mid)
		if log_mixer_mode=="binary_lin": log_mixer = C_Linear_Binary_Linear(dim_in,dim_mid)
		if log_mixer_mode=="sigmoid"   : log_mixer = C_Linear_Sigmoid(dim_in,dim_mid)
		# if log_mixer_mode=="ternary_sig" : log_mixer = C_Linear_Ternary_Sigmoid(dim_in,dim_mid)
		# if log_mixer_mode=="3sigmoid"  : log_mixer = C_Linear_Sigmoid_Ternary(dim_in,dim_mid)
		###
		if out_mixer_mode=="linear"    : out_mixer = C_Linear(dim_mid,dim_out,bias=False)
		if out_mixer_mode=="2softmax"  : out_mixer = C_Linear_2_Softmax(dim_mid,dim_out)
		if out_mixer_mode=="3softmax"  : out_mixer = C_Linear_3_Softmax(dim_mid,dim_out)
		if out_mixer_mode=="binary_sig": out_mixer = C_Linear_Binary_Sigmoid(dim_mid,dim_out)
		if out_mixer_mode=="binary_lin": out_mixer = C_Linear_Binary_Linear(dim_mid,dim_out)
		if out_mixer_mode=="sigmoid"   : out_mixer = C_Linear_Sigmoid(dim_mid,dim_out)
		# if out_mixer_mode=="ternary_sig" : out_mixer = C_Linear_Ternary_Sigmoid(dim_mid,dim_out)
		# if out_mixer_mode=="3sigmoid": out_mixer = C_Linear_Sigmoid_Ternary(dim_mid,dim_out)
		self.model = nn.Sequential(
			ToComplex(),C_Log(),log_mixer,Exp(),ToReal(),out_mixer,
			# ToComplex(),C_Log(),log_mixer,Exp(),out_mixer,ToReal(),
		)
	def forward(self,A,B):
		# return self.model(torch.hstack((A,B)))
		AB = torch.hstack((self.flatten(A),self.flatten(B)))
		return self.model(AB).view(-1,self.dim_A[0],self.dim_B[1])


# dim=2
# dim_A = [dim,dim]
# dim_B = [dim,dim]
# log_mixer_mode="linear"
# out_mixer_mode="linear"
# width=2
# mult = Mul_Log([dim,dim],[dim,dim],width=width,log_mixer_mode=log_mixer_mode,out_mixer_mode=out_mixer_mode)
# name = f"dim_A={dim_A}-dim_A={dim_B}-width={width}-log_mixer={log_mixer_mode}-out_mixer{out_mixer_mode}"
# trainer_square(name,mult,dim,N=2**14,epochs=2**12,plot=False)
# plt.clf()
# plt.tight_layout()
# plt.scatter(range(len(mult.hist)),np.log(mult.hist),s=.05,label=f"{log_mixer_mode} → {out_mixer_mode}")
# plt.show()




# dim_in=64;dim_out=64
# dim_A=[dim_in,dim_out]
# dim_B=[dim_out,1]

def helper_error(name,mult,batch_size=2**16,clamp=10):
	os.makedirs(f"out/extra/",exist_ok=True)
	A = torch.randn(batch_size,*(mult.dim_A),device=device)
	B = torch.randn(batch_size,*(mult.dim_B),device=device)
	truth = A@B
	found = mult(A,B)
	dot = (truth*found).sum(2).sum(1) / ((truth**2).sum(2).sum(1))
	dot = dot.cpu().detach()
	dot=dot.clamp(-clamp,clamp)
	# plt.hist(dot,512,density=True,alpha=0.5,label=name)
	hihi = np.histogram(dot,64,density=True)
	plt.plot([(a+b)/2 for a,b in zip(hihi[1][1:],hihi[1][:-1])], hihi[0], label=name)
	plt.yscale("log")

# "ternary"
# modes = ["linear", "binary", "sigmoid", "2softmax", "3softmax", "3sigmoid",]
modes = ["linear","2softmax","3softmax","binary_sig","binary_lin","sigmoid",]

# make sure we don't lose progress with a crash
def helper4(dim,epochs=2**12,batch_size=2**12,modes1=modes,modes2=modes,mults={},vector_mode=False):
	width=dim
	dim_A = [dim,dim]
	dim_B = [dim,dim]
	if vector_mode: dim_B = [dim,1]
	lr0=-2
	lr1=-4
	 # epochs=2**3
	plt.clf()
	plt.ioff()
	for m in modes1:
		if not m in mults: mults[m]={}
		for n in modes2:
			if vector_mode: name = f"complex_mat{dim}mat{dim}-width={width}-logmixer={m}-outmixer={n}-1d"
			else          : name = f"complex_mat{dim}mat{dim}-width={width}-logmixer={m}-outmixer={n}"
			print(f"{name}")
			if not n in mults[m]:
				mult = Mul_Log(dim_A,dim_B,width=width,log_mixer_mode=m,out_mixer_mode=n).to(device)
				# A= torch.randn(3,*dim_A,device=device)
				# B= torch.randn(3,*dim_B,device=device)
				# mult(A,B)
				mult = trainer(mult,batch_size=batch_size,epochs=epochs,lr0=lr0,lr1=lr1)
				mults[m][n] = mult
				mult = mults[m][n].to(device)
			torch.cuda.empty_cache()
	# return mults

def helper4a(dim,mults,vector_mode,modes1=modes,modes2=modes):
	width=dim
	plt.ioff()
	for m in modes1:
		for n in modes2:
			mult = mults[m][n].to(device)
			if vector_mode: name = f"complex_mat{dim}mat{dim}-width={mult.width}-logmixer={m}-outmixer={n}-1d"
			else          : name = f"complex_mat{dim}mat{dim}-width={mult.width}-logmixer={m}-outmixer={n}"
			print(f"{name}")
			plt.close("all");plt.figure()
			plt.scatter(range(len(mult.hist)),np.log(mult.hist),s=.05,label=name)
			plt.xscale('log')
			plt.ylabel(f"log(MSE)")
			plt.xlabel(f"epoch")
			plt.legend()
			plt.tight_layout()
			plt.grid()
			plt.savefig(f"out/{name}_loss.png")
			viz_error(name,mult,batch_size=2**14)
			mult.to("cpu") # free gpu memory
			torch.cuda.empty_cache()

def helper4b(dim,mults,ylim,modes=modes,name_extra=""):
	os.makedirs(f"out/lin/",exist_ok=True)
	for m in modes:
		plt.clf()
		# plt.figure()
		for n in modes:
			mult = mults[m][n]
			mult.to(device)
			name = f"mat{dim} width={mult.width} {m} → {n}"
			plt.scatter(range(len(mult.hist)),np.log(mult.hist),s=1,label=name)
			mult.to("cpu") # free gpu memory
		plt.legend()
		os.makedirs(f"out/",exist_ok=True)
		plt.ylim(ylim)
		plt.tight_layout()
		plt.savefig(f"out/lin/complex_mat{dim}mat{dim}-width={mult.width}-logmixer={m}{name_extra}.png")
	for n in modes:
		plt.clf()
		for m in modes:
			mult=mults[m][n]
			mult.to(device)
			name = f"mat{dim} width={mult.width} {m} → {n}"
			plt.scatter(range(len(mult.hist)),np.log(mult.hist),s=1,label=name)
			mult.to("cpu") # free gpu memory
		plt.legend()
		plt.ylim(ylim)
		plt.tight_layout()
		plt.savefig(f"out/lin/complex_mat{dim}mat{dim}-width={mult.width}-outmixer={n}{name_extra}.png")

def helper4c(dim,mults,modes=modes,name_extra="",clamp=10,batch_size=2**16):
	os.makedirs(f"out/dot/",exist_ok=True)
	for m in modes:
		plt.close("all");plt.figure()
		for n in modes:
			mult = mults[m][n]
			mult.to(device)
			# name = f"complex_mat{dim}mat{dim}-width={width}-logmixer={m}-outmixer={n}"
			name = f"mat{dim} width={mult.width} {m} → {n}"
			helper_error(name,mult,clamp=clamp,batch_size=batch_size)
			mult.to("cpu") # free gpu memory
			# plt.scatter(range(len(mult.hist)),np.log(mult.hist),s=1,label=f"{m} → {n}")
		plt.plot([1,1],[0,1])
		plt.xlim([-clamp,clamp])
		plt.grid()
		plt.legend(loc="upper left")
		plt.tight_layout()
		os.makedirs(f"out/",exist_ok=True)
		plt.savefig(f"out/dot/complex_mat{dim}mat{dim}-width={mult.width}-logmixer={m}_dot{name_extra}.png")
	for n in modes:
		plt.close("all");plt.figure()
		for m in modes:
			mult = mults[m][n]
			mult.to(device)
			# name = f"complex_mat{dim}mat{dim}-width={width}-logmixer={m}-outmixer={n}"
			name = f"mat{dim} width={mult.width} {m} → {n}"
			helper_error(name,mult,clamp=clamp,batch_size=batch_size)
			mult.to("cpu") # free gpu memory
			# plt.scatter(range(len(mult.hist)),np.log(mult.hist),s=1,label=f"{m} → {n}")
		plt.plot([1,1],[0,1])
		plt.xlim([-clamp,clamp])
		plt.grid()
		plt.legend(loc="upper left")
		plt.tight_layout()
		os.makedirs(f"out/",exist_ok=True)
		plt.savefig(f"out/dot/complex_mat{dim}mat{dim}-width={mult.width}-outmixer={n}_dot{name_extra}.png")

modes2 = ["linear", "binary", "sigmoid", "2softmax", "3softmax",]

# mults2={}
helper4(2,mults=mults2)
helper4a(2,mults2,vector_mode=False)
helper4b(2,mults2,ylim=[-7,2])
helper4c(2,mults2,clamp=5)

# mults3={}
helper4(3,mults=mults3)
helper4a(3,mults2,vector_mode=False)
helper4b(3,mults3,ylim=[-7,2])
helper4c(3,mults3,clamp=5)

# mults8={}
helper4(8,mults=mults8,epochs=2**12,batch_size=2**10)
# helper4(8,mults=mults8)
helper4a(8,mults8,vector_mode=False)
helper4b(8,mults8,ylim=[-6,5])
helper4c(8,mults8,clamp=5)
# helper4b(8,mults8,modes2,"_without3sig")
# helper4c(8,mults8,modes2,"_without3sig")


# mult={}
# mults16 = helper4(16,epochs=2**12,batch_size=2**10)
# # mults16 = helper4(16)
# helper4b(16,mults16)
# helper4c(16,mults16)











# mults1d_2={}
# helper4(2,mults=mults1d_2,vector_mode=True)
# helper4a(2,mults1d_2,vector_mode=True)
# helper4b(2,mults1d_2,ylim=[-7,2],name_extra="-1d")
# helper4c(2,mults1d_2,clamp=5,name_extra="-1d")


mults1d_2 = mult1d_2





# mults1d_3={}
helper4(3,mults=mults1d_3,vector_mode=True)
helper4a(3,mults1d_3,vector_mode=True)
helper4b(3,mults1d_3,ylim=[-7,2],name_extra="-1d")
helper4c(3,mults1d_3,clamp=5,name_extra="-1d")

# mults1d_16={}
# helper4(16,mults=mults1d_16,vector_mode=True)
helper4(16,mults=mults1d_16,vector_mode=True,epochs=2**12,batch_size=2**9)
helper4a(16,mults1d_16,vector_mode=True)
helper4b(16,mults1d_16,ylim=[-7,8],name_extra="-1d")
helper4c(16,mults1d_16,clamp=5,name_extra="-1d",batch_size=2**10)



# complex_mat8mat8-width=8-logmixer=linear-outmixer=linear
# complex_mat8mat8-width=8-logmixer=linear-outmixer=linear tensor(0.)
# complex_mat8mat8-width=8-logmixer=linear-outmixer=binary
# complex_mat8mat8-width=8-logmixer=linear-outmixer=binary tensor(0.4768)
# complex_mat8mat8-width=8-logmixer=linear-outmixer=sigmoid
# complex_mat8mat8-width=8-logmixer=linear-outmixer=sigmoid tensor(0.5029)
# complex_mat8mat8-width=8-logmixer=linear-outmixer=2softmax
# complex_mat8mat8-width=8-logmixer=linear-outmixer=2softmax tensor(0.)
# complex_mat8mat8-width=8-logmixer=linear-outmixer=3softmax
# complex_mat8mat8-width=8-logmixer=linear-outmixer=3softmax tensor(0.)
# complex_mat8mat8-width=8-logmixer=linear-outmixer=3sigmoid
# complex_mat8mat8-width=8-logmixer=linear-outmixer=3sigmoid tensor(0.2188)
# complex_mat8mat8-width=8-logmixer=binary-outmixer=linear
# complex_mat8mat8-width=8-logmixer=binary-outmixer=linear tensor(0.0049)
# complex_mat8mat8-width=8-logmixer=binary-outmixer=binary
# complex_mat8mat8-width=8-logmixer=binary-outmixer=binary tensor(0.2573)
# complex_mat8mat8-width=8-logmixer=binary-outmixer=sigmoid
# complex_mat8mat8-width=8-logmixer=binary-outmixer=sigmoid tensor(0.4986)
# complex_mat8mat8-width=8-logmixer=binary-outmixer=2softmax
# complex_mat8mat8-width=8-logmixer=binary-outmixer=2softmax tensor(0.0001)
# complex_mat8mat8-width=8-logmixer=binary-outmixer=3softmax















# mult = Mul_Log([dim,dim],[dim,dim],width=2,log_mixer_mode="sigmoid") ;trainer_square(mult,dim,N=2**12,epochs=2**11)
# mult = Mul_Log([dim,dim],[dim,dim],width=2,log_mixer_mode="2softmax");trainer_square(mult,dim,N=2**12,epochs=2**11)
# mult = Mul_Log([dim,dim],[dim,dim],width=2,log_mixer_mode="binary")  ;trainer_square(mult,dim,N=2**12,epochs=2**11)
# mult = Mul_Log([dim,dim],[dim,dim],width=2,log_mixer_mode="linear")  ;trainer_square(mult,dim,N=2**12,epochs=2**11)

plt.clf()
viz_error(mult,dim,N=2**14)

# mult = Mul_Log([dim,dim],[dim,dim],width=1);trainer_square(mult,dim,N=2**12,epochs=2**11)
# mult = Mul_Log([dim,dim],[dim,dim],width=2);trainer_square(mult,dim,N=2**12,epochs=2**11)
# mult = Mul_Log([dim,dim],[dim,dim],width=3);trainer_square(mult,dim,N=2**12,epochs=2**11)



# dim=2


# trainer_square(MatMul(dim   ),dim,N=2**15,epochs=2**12)

# trainer_square(MatMulb(dim,1 ,0),dim,N=2**15,epochs=2**14)
# trainer_square(MatMulb(dim,2 ,0),dim,N=2**15,epochs=2**14)
# trainer_square(MatMulb(dim,3 ,0),dim,N=2**15,epochs=2**14)
# trainer_square(MatMulb(dim,5 ,0),dim,N=2**15,epochs=2**12)
# trainer_square(MatMulb(dim,10,0),dim,N=2**15,epochs=2**12)
# trainer_square(MatMulb(dim,20,0),dim,N=2**15,epochs=2**12)

# trainer_square(MatMulb(dim,1 ),dim,N=2**15,epochs=2**14)
# trainer_square(MatMulb(dim,2 ),dim,N=2**15,epochs=2**14)
# trainer_square(MatMulb(dim,3 ),dim,N=2**15,epochs=2**14)
# trainer_square(MatMulb(dim,5 ),dim,N=2**15,epochs=2**14)
# trainer_square(MatMulb(dim,10),dim,N=2**15,epochs=2**14)
# trainer_square(MatMulb(dim,20),dim,N=2**15,epochs=2**14)







































class MatLogDeep(nn.Module):
	def __init__(self,dim_A,dim_B,width=1,depth=5,log_mixer_mode="binary",out_mixer_mode="2softmax"):
		super().__init__()
		if not dim_A[1]==dim_B[0]: print("ERROR INCOMPATIBILITY DETECTED")
		self.dim_A = dim_A
		self.dim_B = dim_B
		self.width=width
		self.flatten = nn.Flatten()
		dim_in  = dim_A[0]*dim_A[1] + dim_B[0]*dim_B[1]
		dim_mid = int(width*dim_in)
		dim_out = dim_A[0]*dim_B[1]
		def log_mixer(dim_i=dim_in,dim_o=dim_mid):
			if log_mixer_mode=="linear"    : return C_Linear(dim_i,dim_o,bias=False)
			if log_mixer_mode=="2softmax"  : return C_Linear_2_Softmax(dim_i,dim_o)
			if log_mixer_mode=="3softmax"  : return C_Linear_3_Softmax(dim_i,dim_o)
			if log_mixer_mode=="binary_sig": return C_Linear_Binary_Sigmoid(dim_i,dim_o)
			if log_mixer_mode=="binary_lin": return C_Linear_Binary_Linear(dim_i,dim_o)
			if log_mixer_mode=="sigmoid"   : return C_Linear_Sigmoid(dim_i,dim_o)
			# if log_mixer_mode=="ternary_sig" : return C_Linear_Ternary_Sigmoid(dim_i,dim_o)
			# if log_mixer_mode=="3sigmoid"  : return C_Linear_Sigmoid_Ternary(dim_i,dim_o)
		###
		def out_mixer(dim_i=dim_mid,dim_o=dim_out):
			if out_mixer_mode=="linear"    : return C_Linear(dim_i,dim_o,bias=False)
			if out_mixer_mode=="2softmax"  : return C_Linear_2_Softmax(dim_i,dim_o)
			if out_mixer_mode=="3softmax"  : return C_Linear_3_Softmax(dim_i,dim_o)
			if out_mixer_mode=="binary_sig": return C_Linear_Binary_Sigmoid(dim_i,dim_o)
			if out_mixer_mode=="binary_lin": return C_Linear_Binary_Linear(dim_i,dim_o)
			if out_mixer_mode=="sigmoid"   : return C_Linear_Sigmoid(dim_i,dim_o)
			# if out_mixer_mode=="ternary_sig" : return C_Linear_Ternary_Sigmoid(dim_i,dim_o)
			# if out_mixer_mode=="3sigmoid": return C_Linear_Sigmoid_Ternary(dim_i,dim_o)
		self.model = nn.Sequential(
			ToComplex(),C_Log(),log_mixer(dim_in,dim_mid),Exp(),ToReal(),out_mixer(dim_mid,dim_mid),
			*[
				nn.Sequential(ToComplex(),C_Log(),log_mixer(dim_mid,dim_mid),Exp(),ToReal(),out_mixer(dim_mid,dim_mid),)
				for d in range(depth)
			],
			ToComplex(),C_Log(),log_mixer(dim_mid,dim_mid),Exp(),ToReal(),out_mixer(dim_mid,dim_out),
		)
	def forward(self,A,B):
		AB = torch.hstack((self.flatten(A),self.flatten(B)))
		return self.model(AB).view(-1,self.dim_A[0],self.dim_B[1])

def helper5(dim,width,depth,epochs=2**12,batch_size=2**12,modes1=modes,modes2=modes,mults={}):
	dim_A = [dim,dim]
	dim_B = [dim,1]
	lr0=-2
	lr1=-4
	 # epochs=2**3
	plt.clf()
	plt.ioff()
	for m in modes1:
		if not m in mults: mults[m]={}
		for n in modes2:
			name = f"complex_deep_mat{dim}vec{dim}-width={width}-depth{depth}-logmixer={m}-outmixer={n}"
			print(f"{name}")
			if not n in mults[m]:
				mult = MatLogDeep(dim_A,dim_B,width=width,depth=depth,log_mixer_mode=m,out_mixer_mode=n).to(device)
				mult.name = name
				# A= torch.randn(3,*dim_A,device=device)
				# B= torch.randn(3,*dim_B,device=device)
				# mult(A,B)
				mult = trainer(mult,batch_size=batch_size,epochs=epochs,lr0=lr0,lr1=lr1)
				mults[m][n] = mult.to("cpu")
			torch.cuda.empty_cache()

# mults1d_2_cheap={}
helper5(2,width=1,depth=5,mults=mults1d_2_cheap,batch_size=2**10,epochs=2**11)


















class MatLogDeep2(nn.Module):
	def __init__(self,dim_A,dim_B,width=1,depth=5,log_mixer_mode="binary",out_mixer_mode="2softmax"):
		super().__init__()
		if not dim_A[1]==dim_B[0]: print("ERROR INCOMPATIBILITY DETECTED")
		self.dim_A = dim_A
		self.dim_B = dim_B
		self.width=width
		self.flatten = nn.Flatten()
		dim_in  = dim_A[0]*dim_A[1] + dim_B[0]*dim_B[1]
		dim_mid = int(width*dim_in)
		dim_out = dim_A[0]*dim_B[1]
		def log_mixer(dim_i=dim_in,dim_o=dim_mid):
			if log_mixer_mode=="linear"    : return C_Linear(dim_i,dim_o,bias=False)
			if log_mixer_mode=="2softmax"  : return C_Linear_2_Softmax(dim_i,dim_o)
			if log_mixer_mode=="3softmax"  : return C_Linear_3_Softmax(dim_i,dim_o)
			if log_mixer_mode=="binary_sig": return C_Linear_Binary_Sigmoid(dim_i,dim_o)
			if log_mixer_mode=="binary_lin": return C_Linear_Binary_Linear(dim_i,dim_o)
			if log_mixer_mode=="sigmoid"   : return C_Linear_Sigmoid(dim_i,dim_o)
			# if log_mixer_mode=="ternary_sig" : return C_Linear_Ternary_Sigmoid(dim_i,dim_o)
			# if log_mixer_mode=="3sigmoid"  : return C_Linear_Sigmoid_Ternary(dim_i,dim_o)
		###
		def out_mixer(dim_i=dim_mid,dim_o=dim_out):
			if out_mixer_mode=="linear"    : return C_Linear(dim_i,dim_o,bias=False)
			if out_mixer_mode=="2softmax"  : return C_Linear_2_Softmax(dim_i,dim_o)
			if out_mixer_mode=="3softmax"  : return C_Linear_3_Softmax(dim_i,dim_o)
			if out_mixer_mode=="binary_sig": return C_Linear_Binary_Sigmoid(dim_i,dim_o)
			if out_mixer_mode=="binary_lin": return C_Linear_Binary_Linear(dim_i,dim_o)
			if out_mixer_mode=="sigmoid"   : return C_Linear_Sigmoid(dim_i,dim_o)
			# if out_mixer_mode=="ternary_sig" : return C_Linear_Ternary_Sigmoid(dim_i,dim_o)
			# if out_mixer_mode=="3sigmoid": return C_Linear_Sigmoid_Ternary(dim_i,dim_o)
		self.model = nn.Sequential(
			ToComplex(),C_Log(),C_Linear(dim_in,dim_mid,bias=False),
			*[nn.Sequential(log_mixer(dim_mid,dim_mid),Exp(),ToReal(),nn.ReLU(),ToComplex(),C_Log()) for d in range(depth)],
			Exp(),ToReal(),C_Linear(dim_mid,dim_out,bias=False),
		)
		# self.model = nn.Sequential(
		# 	ToComplex(),C_Log(),log_mixer(dim_in,dim_mid),Exp(),ToReal(),out_mixer(dim_mid,dim_mid),
		# 	*[nn.Sequential(ToComplex(),C_Log(),log_mixer(dim_mid,dim_mid),Exp(),ToReal(),out_mixer(dim_mid,dim_mid),) for d in range(depth)],
		# 	out_mixer(dim_mid,dim_out),
		# )
	def forward(self,A,B):
		AB = torch.hstack((self.flatten(A),self.flatten(B)))
		return self.model(AB).view(-1,self.dim_A[0],self.dim_B[1])

def helper6(dim,width,depth,epochs=2**12,batch_size=2**12,mults={}):
	dim_A = [dim,dim]
	dim_B = [dim,1]
	lr0=-2
	lr1=-4
	 # epochs=2**3
	plt.clf()
	plt.ioff()
	# for m in ["3softmax"]:
	for m in ["linear"]:
		if not m in mults: mults[m]={}
		for n in ["linear"]:
			name = f"complex_deep2_mat{dim}vec{dim}-width={width}-depth{depth}-logmixer={m}-outmixer={n}"
			print(f"{name}")
			if not n in mults[m]:
				mult = MatLogDeep2(dim_A,dim_B,width=width,depth=depth,log_mixer_mode=m,out_mixer_mode=n).to(device)
				mult.name = name
				# A= torch.randn(3,*dim_A,device=device)
				# B= torch.randn(3,*dim_B,device=device)
				# mult(A,B)
				mult = trainer(mult,batch_size=batch_size,epochs=epochs,lr0=lr0,lr1=lr1)
				mults[m][n] = mult.to("cpu")
			torch.cuda.empty_cache()

mults1d2_2_cheap={}
# helper6(2,width=3,depth=0,mults=mults1d2_2_cheap,batch_size=2**10,epochs=2**11)
# helper6(2,width=2,depth=1,mults=mults1d2_2_cheap,batch_size=2**10,epochs=2**11)
helper6(2,width=2,depth=0,mults=mults1d2_2_cheap,batch_size=2**10,epochs=2**11)









